<?php

/* report.php - SynApp2 PDF report generator
**
   +----------------------------------------------------------------------+
   | SynApp2 Version 1                                                    |
   +----------------------------------------------------------------------+
   | Copyright (c) 2007 - 2010 Richard Howell. All rights reserved.       |
   +----------------------------------------------------------------------+
   | This source file is subject to version 1.01 of the SynApp2 license,  |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.synapp2.org/license/1_01.txt                              |
   | If you did not receive a copy of the SynApp2 license and are unable  |
   | to obtain it through the world-wide-web, please send a note to       |
   | license@synapp2.org so we can mail you a copy immediately.           |
   +----------------------------------------------------------------------+
   | Authors: Richard Howell                                              |
   |                                                                      |
   +----------------------------------------------------------------------+
**
** http://www.synapp2.org
**
** $Id: report.php,v 1.10 2010/07/19 01:30:44 richard Exp $
*/

require('./3rdparty/fpdf.php');
require('./3rdparty/mc_table.php');

class report extends PDF_MC_Table // extends FPDF
{
    var $m_omit = array();
    var $m_col_order_map = null;
    var $m_hdgs = array();
    var $m_hdgs_align = array();
    var $m_align = array();
    var $m_width = array();
    var $m_detail_summary = null;
    var $m_cell_border = false;
    var $m_row_shade = false;
    var $m_row_shaded = false;

    ////

    function report($orientation = 'l', $unit = 'pt', $format = 'letter')
    {
        global $g_custom;

        $this->FPDF($orientation, $unit, $format);
        $this->SetDisplayMode('default', 'default');
        $this->SetCreator('SynApp2 - version ' . get_version(true));

        $this->SetTitle($g_custom->get_form_title(RFORM));
        $this->m_cell_border = $g_custom->is_rpt_cell_border();
        $this->m_row_shade = $g_custom->is_rpt_row_shade();
    }

    function is_omit_col($col_name)
    {
        return !empty($this->m_omit[$col_name]);
    }

    function is_cell_bordered()
    {
        return $this->m_cell_border;
    }

    function is_row_shaded()
    {
        return $this->m_row_shade && $this->m_row_shaded;
    }

    function toggle_is_row_shaded()
    {
        $this->m_row_shaded = !$this->m_row_shaded;
    }

    function reset_is_row_shaded()
    {
        $this->m_row_shaded = false;
    }

    function em($mult = 1)
    {
        if (!is_numeric($mult) || $mult <= 0)
        {
            $mult = 1;
        }

        return $this->GetStringWidth('M') * ((float)$mult);
    }

    function get_col_name_iter(&$rec)
    {
        return $this->m_col_order_map ? array_values($this->m_col_order_map) : array_keys($rec) ;
    }

    function auto_widths(&$data, $expand_to_fit = true)
    {
        global $g_custom;

        $this->m_width = array();

        if (!empty($data))
        {
            $available = round(($this->w - $this->lMargin - $this->rMargin) / $this->em()); // line width
            $minsize = 0;
            $maxsize = 0;
            $colsize = array(); // starting size estimates
            $required = 0; // total of size estimates
            $deficit = 0;

            foreach ($data as $row)
            {
                $col_count = count($row) - count($this->m_omit);

                if (!$col_count)
                {
                    break; // TODO: TBD?? message??
                }

                $minsize = max(1, min(5, (int)($available / $col_count)));
                $maxsize = min(40, $minsize * 8);

                foreach ($row as $n => $v)
                {
                    if (!$this->is_omit_col($n))
                    {
                        $size = max($minsize, min($maxsize, strlen($v)));

                        if (!isset($colsize[$n]))
                        {
                            $colsize[$n] = $size;
                        }
                        else
                        {
                            $colsize[$n] = max($size, $colsize[$n]);
                        }
                    }
                }
            }

            $required = array_sum($colsize);
            $flexible = $required;

            foreach ($colsize as $n => $size)
            {
                $s = $g_custom->get_form_col_size(RFORM, $n);

                if ($s)
                {
                    $size = $colsize[$n];
                    $colsize[$n] = $s;
                    $flexible -= $size;
                    $delta = $s - $size;
                    $required += $delta;
                }
            }

            $deficit = $required - $available;

            $col_name_iter = $this->get_col_name_iter($data[0]);

            foreach ($col_name_iter as $n)
            {
                if (!$this->is_omit_col($n))
                {
                    $s = $g_custom->get_form_col_size(RFORM, $n);

                    if ($s)
                    {
                        $this->m_width[$n] = $this->em($s);
                    }
                    else
                    {
                        $size = $colsize[$n];
                        $delta = $flexible ? (($size / $flexible) * $deficit) : 0 ;
                        $this->m_width[$n] = $this->em($size - ($expand_to_fit ? $delta : max(0, $delta)));
                    }
                }
            }
        }
    }

    function prepare_cols(&$data)
    {
        global $g_action, $g_custom, $g_schema, $g_query;

        $this->m_omit = array();
        $this->m_col_order_map = null;
        $this->m_hdgs = array();
        $this->m_hdgs_align = array();
        $this->m_align = array();
        $this->m_width = array();

        if (!empty($data))
        {
            $rec = $data[0];

            $this->m_col_order_map = $g_custom->get_form_col_order_map(RFORM, array_keys($rec));

            $map = $g_schema->get_col_display_name_map($g_query->get_table());

            $col_name_iter = $this->get_col_name_iter($rec);

            foreach ($col_name_iter as $col_name)
            {
                $col_value = $rec[$col_name];
                $table_name = $g_query->get_table();

                if ($g_schema->is_pk($table_name, $col_name) ||
                    $g_schema->is_fk($table_name, $col_name) ||
                    $g_custom->is_form_omit_col(RFORM, $col_name, $g_action->get_qid()))
                {
                    $this->m_omit[$col_name] = true;
                }
                else
                {
                    $display_name = fmt_display_name($col_name);

                    if (isset($map[$col_name]))
                    {
                        $display_name = $map[$col_name];
                    }
                    else if (strpos($col_name, "{$table_name}_") === 0) // TODO: encapsulate
                    {
                        $fk_name = str_replace("{$table_name}_", '', $col_name); // TODO: encapsulate

                        if (isset($map[$fk_name]))
                        {
                            $display_name = $map[$fk_name];
                        }
                    }

                    $this->m_hdgs[$col_name] = $display_name;
                    $this->m_hdgs_align[$col_name] = ALIGN_C;
                    $this->m_align[$col_name] = $g_custom->get_form_col_align(RFORM, $col_name, (is_numeric($col_value) ? ALIGN_R : ALIGN_L));
                }
            }

            $this->auto_widths($data);

            $this->m_detail_summary = new detail_summary($g_custom, $data);
        }
    }

    function get_col_values($rec)
    {
        global $g_custom;

        static $col_name_iter = null;

        if (!$col_name_iter)
        {
            $col_name_iter = $this->get_col_name_iter($rec);
        }

        $col_values = array();

        foreach ($col_name_iter as $n)
        {
            if (!$this->is_omit_col($n))
            {
                $col_values[] = isset($rec[$n]) ? $g_custom->form_format_col(RFORM, $n, $rec[$n]) : '' ;
            }
        }

        return $col_values;
    }

    function Header()
    {
        $this->reset_is_row_shaded(); // first row always not shaded

        $this->SetY($this->bMargin / 2); // down from the top
        $font_size = $this->FontSizePt;
        $this->SetFontSize($font_size * 1.25);
        $this->Cell(0, 0, $this->title, 0, 0, ALIGN_C);
        $this->Ln(2 * $this->em());
        $this->SetFontSize($font_size);

        if (!empty($this->m_hdgs))
        {
            $this->SetAligns(array_values($this->m_hdgs_align));
            $this->Row(array_values($this->m_hdgs), true, true);
            $this->Ln(0.25 * $this->em());
            $this->SetAligns(array_values($this->m_align));
        }
    }

    function Body(&$data)
    {
        if (!empty($data))
        {
            $this->prepare_cols($data);

            $this->SetAligns(array_values($this->m_align));
            $this->SetWidths(array_values($this->m_width));

            $this->AddPage();

            foreach ($data as $rec)
            {
                if ($summary_row = $this->m_detail_summary->get_summary_row($rec))
                {
                    $this->Ln(0.25 * $this->em());
                    $this->Row($this->get_col_values($summary_row), false, !$this->m_row_shade || $this->is_row_shaded());
                    $this->Ln(0.25 * $this->em());
                    $this->toggle_is_row_shaded();
                    $this->m_detail_summary->suppress_col_shift_left();
                }

                $this->Row($this->get_col_values($rec), $this->is_cell_bordered(), $this->is_row_shaded(), $this->m_detail_summary->get_suppress_col());
                $this->toggle_is_row_shaded();
                $this->m_detail_summary->suppress_col_shift_right();

                if ($summary_row = $this->m_detail_summary->get_final_row($rec))
                {
                    $this->Ln(0.25 * $this->em());
                    $this->Row($this->get_col_values($summary_row), false, !$this->m_row_shade || $this->is_row_shaded());
                    $this->Ln(0.25 * $this->em());
                    $this->toggle_is_row_shaded();
                }

                if ($summary_row = $this->m_detail_summary->get_grand_row($rec))
                {
                    $this->Ln(0.25 * $this->em());
                    $this->Row($this->get_col_values($summary_row), false, !$this->m_row_shade || $this->is_row_shaded());
                    $this->Ln(0.25 * $this->em());
                    $this->toggle_is_row_shaded();
                }
            }
        }
    }

    function Footer()
    {
        if (empty($this->AliasNbPages))
        {
            $this->AliasNbPages();
        }

        $this->SetY($this->bMargin / -2); // up from the bottom
        $this->Cell(0, 0, 'Page ' . $this->PageNo() . ' of ' . $this->AliasNbPages, 0, 0, 'C');
    }
};

////////

class detail_summary
{
    var $m_custom = null;

    var $m_rows_remaining = 0;
    var $m_col_names = array();
    var $m_detail_summary_cols = array();
    var $m_isuppress_cols = 0;
    var $m_suppress_cols = array();

    var $m_group_stat = array();
    var $m_grand_stat = array();

    function detail_summary(&$custom, &$data)
    {
        $this->m_custom = &$custom;
        $this->build($data);
    }

    function build(&$data)
    {
        if (!empty($data))
        {
            $this->m_rows_remaining = count($data);
            $this->m_col_names = array_keys($data[0]);
        }

        $this->m_detail_summary_cols = $this->m_custom->get_detail_summary_cols();

        $this->m_stage_suppress_col_shift_right = 0;
        $this->m_isuppress_cols = 0;
        $this->m_suppress_cols = array(0);

        if (!empty($this->m_detail_summary_cols))
        {
            foreach ($this->m_detail_summary_cols as $summary_col)
            {
                $this->m_suppress_cols[] = isset($summary_col[2]) ? (int)$summary_col[2] : 0 ;
            }

            sort($this->m_suppress_cols);
        }
    }

    function get_suppress_col()
    {
        return isset($this->m_suppress_cols[$this->m_isuppress_cols]) ? $this->m_suppress_cols[$this->m_isuppress_cols] : 0 ;
    }

    function stage_suppress_col_shift_right()
    {
        $this->m_stage_suppress_col_shift_right += 1;
    }

    function suppress_col_shift_right()
    {
        while ($this->m_stage_suppress_col_shift_right--)
        {
            if (isset($this->m_suppress_cols[1 + $this->m_isuppress_cols]))
            {
                $this->m_isuppress_cols += 1;
            }
            else
            {
                break; // for safety
            }
        }

        $this->m_stage_suppress_col_shift_right = 0;
    }

    function suppress_col_shift_left()
    {
        if ($this->m_isuppress_cols > 0)
        {
            $this->m_isuppress_cols -= 1;
        }
    }

    function reset_summary_row(&$summary_row)
    {
        if (!empty($this->m_col_names))
        {
            foreach ($this->m_col_names as $col_name)
            {
                $summary_row[$col_name] = '';
            }
        }
    }

    function update_col_stat(&$col_stat, &$rec, $col_name, $group_col_name = null)
    {
        $col_stat['count'] += 1;

        if (is_numeric($rec[$col_name]))
        {
            $col_value = 0 + $rec[$col_name];

            $col_stat['sum'] += $col_value;
            $col_stat['ave'] = $col_stat['sum'] / $col_stat['count'];

            if ($col_value < $col_stat['min'] || $col_stat['min'] === null)
            {
                $col_stat['min'] = $col_value;
            }

            if ($col_value > $col_stat['max'] || $col_stat['max'] === null)
            {
                $col_stat['max'] = $col_value;
            }
        }
        else
        {
            $col_stat['sum'] = '';
        }

        if ($col_stat['last'] === null && isset($rec[$group_col_name]))
        {
            $col_stat['last'] = $rec[$group_col_name];
            $this->stage_suppress_col_shift_right();
        }
    }

    function make_col_stat()
    {
        return array('count'=>0, 'sum'=>0, 'ave'=>0, 'min'=>null, 'max'=>null, 'last'=>null);
    }

    function tally_col_stat(&$rec, $col_name, $group_col_name = null)
    {
        $stat_rec = null;

        if ($col_name && isset($rec[$col_name]) /* && is_numeric($rec[$col_name]) */) // TBD: is_numeric() test not-needed or wrong place??
        {
            if (!isset($this->m_grand_stat[$col_name]))
            {
                $this->m_grand_stat[$col_name] = $this->make_col_stat();
            }

            $this->update_col_stat($this->m_grand_stat[$col_name], $rec, $col_name);

            ////////

            if ($group_col_name && isset($rec[$group_col_name]))
            {
                if (!isset($this->m_group_stat[$col_name][$group_col_name]))
                {
                    $this->m_group_stat[$col_name][$group_col_name] = $this->make_col_stat();
                }

                $col_stat = &$this->m_group_stat[$col_name][$group_col_name];

                if ($col_stat['last'] && $col_stat['last'] != $rec[$group_col_name])
                {
                    $stat_rec = $this->m_group_stat[$col_name][$group_col_name]; // copy by value

                    $this->m_group_stat[$col_name][$group_col_name] = $this->make_col_stat();

                    $col_stat = &$this->m_group_stat[$col_name][$group_col_name];
                }

                $this->update_col_stat($col_stat, $rec, $col_name, $group_col_name);
            }
        }

        return $stat_rec;
    }

    function get_group_stat($col_name, $group_col_name)
    {
        return isset($this->m_group_stat[$col_name][$group_col_name]) ? ($this->m_group_stat[$col_name][$group_col_name]) : null ;
    }

    function get_grand_stat($col_name)
    {
        return isset($this->m_grand_stat[$col_name]) ? ($this->m_grand_stat[$col_name]) : null ;
    }

    function get_summary_row(&$rec, $selector = null)
    {
        $summary_row = null;

        if (!empty($rec) && !empty($this->m_detail_summary_cols))
        {
            foreach ($this->m_detail_summary_cols as $summary_col)
            {
                if (empty($summary_col[0]) || empty($summary_col[1]))
                {
                    continue;
                }

                $col_name = $summary_col[0];
                $group_col_name = $summary_col[1];

                $col_stat = null;

                if (!$selector)
                {
                    $col_stat = $this->tally_col_stat($rec, $col_name, $group_col_name);
                }
                else if (!$this->m_rows_remaining)
                {
                    if ($selector == 'final')
                    {
                        $col_stat = $this->get_group_stat($col_name, $group_col_name);
                    }
                    else if ($selector == 'grand')
                    {
                        $col_stat = $this->get_grand_stat($col_name);
                    }
                }

                if ($col_stat)
                {
                    if (!$summary_row)
                    {
                        $this->reset_summary_row($summary_row);
                    }

                    $summary_row[$col_name] = $col_stat['sum'];
                }
            }

            if ($this->m_rows_remaining)
            {
                $this->m_rows_remaining--;
            }
        }

        return $summary_row;
    }

    function get_final_row(&$rec)
    {
        return !$this->m_rows_remaining ? $this->get_summary_row($rec, 'final') : null ;
    }

    function get_grand_row(&$rec)
    {
        return !$this->m_rows_remaining ? $this->get_summary_row($rec, 'grand') : null ;
    }
};

////

?>
